import Flutter
import UIKit
import SlangRetailAssistant

public class SwiftSlangRetailAssistantPlugin: NSObject, FlutterPlugin {
    
    private static let methodChannelName = "plugins.flutter.io/slang_retail_assistant";
    private static let configAssistantId = "assistantId";
    private static let configAPIKey = "apiKey";
    private static let configLocale = "defaultLocale";
    private static let configRequestLocales = "requestedLocales";
    private static let configTriggerPosition = "triggerPosition";
    private static let configCustomTriggerEnable = "enableCustomTrigger";
    private static let environment = "environment";

    private var methodChannel:FlutterMethodChannel?
    private var searchUserJourney:SearchUserJourney?
    private var navigationUserJourney:NavigationUserJourney?

    init(_ methodChannel:FlutterMethodChannel) {
        self.methodChannel = methodChannel
    }
    
    public static func register(with registrar: FlutterPluginRegistrar) {
        let channel = FlutterMethodChannel(name: methodChannelName, binaryMessenger: registrar.messenger())
        let instance = SwiftSlangRetailAssistantPlugin(channel)
        registrar.addMethodCallDelegate(instance, channel: channel)
    }
    
    public func handle(_ call: FlutterMethodCall, result: @escaping FlutterResult) {
        switch call.method {
        
        case "initialize":
            guard let options = call.arguments as? Dictionary<String, Any> else {
                sendAssistantError(
                    SlangFlutterWrapperError(
                        "Unable to parse the required arguments")
                )
                return
            }
            SlangRetailAssistantUtils.setConfigOptions(options)
            let config = AssistantConfiguration.Builder()
                .setAssistantId(
                    SlangRetailAssistantUtils.getStringConfig(
                        SwiftSlangRetailAssistantPlugin.configAssistantId
                    )
                )
                .setAPIKey(
                    SlangRetailAssistantUtils.getStringConfig(
                        SwiftSlangRetailAssistantPlugin.configAPIKey
                    )
                )
                .setDefaultLocale(
                    SlangRetailAssistantLocaleMap.getDefaultLocale(
                        SlangRetailAssistantUtils.getStringConfig(SwiftSlangRetailAssistantPlugin.configLocale)
                    )
                )
                .setRequestedLocales(
                    SlangRetailAssistantLocaleMap.getRequestedLocales(
                        SlangRetailAssistantUtils.getStringListConfig(
                            SwiftSlangRetailAssistantPlugin.configRequestLocales)
                    )
                )
                .enableCustomTrigger(
                    SlangRetailAssistantUtils.getBooleanConfig(
                        SwiftSlangRetailAssistantPlugin.configCustomTriggerEnable)
                )
                .setUIPosition(
                    SlangRetailAssistantUtils
                        .getAssistantUIConfig(
                            SwiftSlangRetailAssistantPlugin.configTriggerPosition)
                )
                .setEnvironment(with: AssistantConfiguration.Environment(
                    rawValue:
                        SlangRetailAssistantUtils.getStringConfig(
                            SwiftSlangRetailAssistantPlugin.environment)) ?? .staging
                )
                .build()
            
            SlangRetailAssistant.initialize(with: config)
            
        case "setAction":
            SlangRetailAssistant.setAction(with: self)
            
        case "setLifecycleObserver":
            SlangRetailAssistant.setLifecycleObserver(with: self)
            
        case "showTrigger":
            guard let window = UIApplication.shared.keyWindow else {return}
            if let rootViewController = window.rootViewController {
                SlangRetailAssistant.getUI().showTrigger(in: rootViewController)
            }
            
        case "hideTrigger":
            SlangRetailAssistant.getUI().hideTrigger()
            
        case "startConversation":
            guard let options = call.arguments as? Dictionary<String, Any> else {
                sendAssistantError(SlangFlutterWrapperError("Unable to parse the required arguments"))
                return
            }
            let userJourney = RetailUserJourney(rawValue: options["userJourney"] as? String ?? "") ?? .search
            guard let window = UIApplication.shared.keyWindow else {return}
            if let rootViewController = window.rootViewController {
                SlangRetailAssistant.startConversation(
                    userJourney,
                    viewController: rootViewController
                )
            }
            
        case "setSearchSuccess":
            searchUserJourney?.setSuccess()
            
        case "setSearchFailure":
            searchUserJourney?.setFailure()
            
        case "setSearchItemNotFound":
            searchUserJourney?.setItemNotFound()
            
        case "setSearchItemOutOfStock":
            searchUserJourney?.setItemOutOfStock()
            
        case "setSearchItemNotSpecified":
            searchUserJourney?.setItemNotSpecified()
            
        case "setSearchNeedItemQuantity":
            searchUserJourney?.setNeedItemQuantity()
            
        case "setSearchNeedDisambiguation":
            searchUserJourney?.setNeedDisambiguation()
            
        case "setSearchStopDisambiguation":
            searchUserJourney?.setStopDisambiguation()
            
        case "notifySearchActionCompleted":
            guard let options = call.arguments as? Dictionary<String, Any> else {
                sendAssistantError(
                    SlangFlutterWrapperError("Unable to parse the required arguments")
                )
                return
            }
            let appStateString = options["appState"] as? String
            let appState = SearchUserJourney.AppState(rawValue: appStateString ?? "") ?? .unsupported
            do {
                try searchUserJourney?.notifyAppState(with: appState)
            }
            catch {
                sendAssistantError(error)
            }
            
        case "setUserJourneyToContinue":
            guard let options = call.arguments as? Dictionary<String, Any> else {
                sendAssistantError(SlangFlutterWrapperError("Unable to parse the required arguments"))
                return
            }
            
            let userJourney
                = RetailUserJourney(rawValue: options["currentUserJourney"] as? String ?? "") ?? .search
            let userJourneyToContinue
                = RetailUserJourney(rawValue: options["userJourneyToContinue"] as? String ?? "") ?? .search
            let isSpoken = options["isSpoken"] as? Bool ?? false
            
            switch userJourney {
            case .search:
                searchUserJourney?
                    .setRetailJourneyToContinue(
                        with: userJourneyToContinue,
                        isSpoken: isSpoken
                    )
                
            case .navigation:
                navigationUserJourney?
                    .setRetailJourneyToContinue(
                        with: userJourneyToContinue,
                        isSpoken: isSpoken
                    )
                
            default:
                break
            }
            
        case "clearUserJourneyToContinue":
            guard let options = call.arguments as? Dictionary<String, Any> else {
                sendAssistantError(SlangFlutterWrapperError("Unable to parse the required arguments"))
                return
            }
            
            let userJourney
                = RetailUserJourney(rawValue: options["userJourney"] as? String ?? "") ?? .search
            
            switch userJourney {
            case .search:
                searchUserJourney?
                    .clearRetailJourneyToContinue()
                
            case .navigation:
                searchUserJourney?
                    .clearRetailJourneyToContinue()
                
            default:
                break
            }
            
        case "setAppDefaultUserJourney":
            guard let options = call.arguments as? Dictionary<String, Any> else {
                sendAssistantError(SlangFlutterWrapperError("Unable to parse the required arguments"))
                return
            }
            
            let userJourney
                = RetailUserJourney(rawValue: options["userJourney"] as? String ?? "") ?? .search
            
            SlangRetailAssistant.setAppDefaultUserJourney(with: userJourney)
            
        case "setAppDefaultSubDomain":
            guard let options = call.arguments as? Dictionary<String, Any> else {
                sendAssistantError(SlangFlutterWrapperError("Unable to parse the required arguments"))
                return
            }
            
            let appSubDomain
                = AssistantSubDomain(rawValue: options["appSubDomain"] as? String ?? "") ?? .grocery
            
            SlangRetailAssistant.setAppDefaultSubDomain(with: appSubDomain)
            
        case "clearSearchUserJourneyContext":
            SearchUserJourney.getContext().clear()
            
        case "disableSearchUserJourneyContext":
            SearchUserJourney.disablePreserveContext()
            
        case "setNavigationSuccess":
            navigationUserJourney?.setNavigationSuccess()
            
        case "setNavigationFailure":
            navigationUserJourney?.setNavigationFailure()

        case "setTargetNotSpecified":
                    navigationUserJourney?.setTargetNotSpecified()
            
        case "notifyNavigationActionCompleted":
            guard let options = call.arguments as? Dictionary<String, Any> else {
                sendAssistantError(
                    SlangFlutterWrapperError("Unable to parse the required arguments")
                )
                return
            }
            let appStateString = options["appState"] as? String
            let appState = NavigationUserJourney.AppState(rawValue: appStateString ?? "") ?? .unsupported
            do {
                try navigationUserJourney?.notifyAppState(with: appState)
            }
            catch {
                sendAssistantError(error)
            }
            
        case "disableNavigationUserJourneyContext":
            NavigationUserJourney.disablePreserveContext()
            
        case "clearNavigationUserJourneyContext":
            NavigationUserJourney.getContext().clear()
            
        default:
            print("no method")
        }
        
    }
    
    private func sendAssistantError(_ error:Error) {
        DispatchQueue.main.async { [weak self] in
            guard let strongSelf = self else {return}
            var params = [String:String]()
            params["type"] = error.localizedDescription
            strongSelf.methodChannel?.invokeMethod(
                "onAssistantError",
                arguments: params
            )
        }
    }
    
}

extension SwiftSlangRetailAssistantPlugin:SlangRetailAssistantAction {
    public func onSearch(with searchInfo: SearchInfo, searchJourney: SearchUserJourney) -> SearchUserJourney.AppState {
        self.searchUserJourney = searchJourney
        DispatchQueue.main.async { [weak self] in
            guard let strongSelf = self else {return}
            strongSelf.methodChannel?.invokeMethod(
                "onSearch",
                arguments: SlangRetailAssistantUtils.mapSearchInfo(searchInfo)
            )
        }
        return .waiting
    }
    
    public func onNavigation(with navigationInfo: NavigationInfo, navigationUserJourney: NavigationUserJourney) -> NavigationUserJourney.AppState {
        self.navigationUserJourney = navigationUserJourney
        DispatchQueue.main.async { [weak self] in
            guard let strongSelf = self else {return}
            strongSelf.methodChannel?.invokeMethod(
                "onNavigation",
                arguments: SlangRetailAssistantUtils.mapNavigationInfo(navigationInfo)
            )
        }
        return .waiting
    }
    
    public func onAssistantError(with error: AssistantError) {
        DispatchQueue.main.async { [weak self] in
            guard let strongSelf = self else {return}
            var params = [String:Any]()
            params["description"] = error.description
            strongSelf.methodChannel?.invokeMethod("onAssistantError", arguments: params)
        }
    }
}


extension SwiftSlangRetailAssistantPlugin:LifecycleObserver {
    
    public func onAssistantInitSuccess() {
        DispatchQueue.main.async { [weak self] in
            guard let strongSelf = self else {return}
            strongSelf.methodChannel?.invokeMethod("onAssistantInitSuccess", arguments: nil)
        }
    }
    
    public func onAssistantInitFailure(with description: String) {
        DispatchQueue.main.async { [weak self] in
            guard let strongSelf = self else {return}
            var params = [String:Any]()
            params["type"] = description
            strongSelf.methodChannel?.invokeMethod("onAssistantInitFailure", arguments: params)
        }
    }
    
    public func onAssistantInvoked() {
        DispatchQueue.main.async { [weak self] in
            guard let strongSelf = self else {return}
            strongSelf.methodChannel?.invokeMethod("onAssistantInvoked", arguments: nil)
        }
    }
    
    public func onAssistantClosed(with cancelled: Bool) {
        DispatchQueue.main.async { [weak self] in
            guard let strongSelf = self else {return}
            var params = [String:Any]()
            params["isCancelled"] = cancelled
            strongSelf.methodChannel?.invokeMethod("onAssistantClosed", arguments: params)
        }
    }
    
    public func onAssistantLocaleChanged(for locale: Locale) {
        DispatchQueue.main.async { [weak self] in
            guard let strongSelf = self else {return}
            var params = [String:Any]()
            params["country"] = locale.regionCode
            params["language"] = locale.languageCode
            strongSelf.methodChannel?.invokeMethod("onAssistantLocaleChanged", arguments: params)
        }
    }
    
    public func onUnrecognisedUtterance(for utterance: String?) {
        DispatchQueue.main.async { [weak self] in
            guard let strongSelf = self else {return}
            var params = [String:Any]()
            params["utterance"] = utterance
            strongSelf.methodChannel?.invokeMethod("onUnrecognisedUtterance", arguments: params)
        }
    }
    
    public func onUtteranceDetected(for utterance: String?) {
        DispatchQueue.main.async { [weak self] in
            guard let strongSelf = self else {return}
            var params = [String:Any]()
            params["utterance"] = utterance
            strongSelf.methodChannel?.invokeMethod("onUtteranceDetected", arguments: params)
        }
    }
    
    public func onOnboardingSuccess(for utterance: String?) {
        DispatchQueue.main.async { [weak self] in
            guard let strongSelf = self else {return}
            strongSelf.methodChannel?.invokeMethod("onOnboardingSuccess", arguments: nil)
        }
    }
    
    public func onOnboardingFailure() {
        DispatchQueue.main.async { [weak self] in
            guard let strongSelf = self else {return}
            strongSelf.methodChannel?.invokeMethod("onOnboardingFailure", arguments: nil)
        }
    }
}
